import logging

from typing import Generic
from typing import TypeVar
from typing import cast
from typing import override

import jwt

from django.contrib.auth.models import AnonymousUser
from requests import HTTPError
from rest_framework.authentication import BaseAuthentication
from rest_framework.exceptions import AuthenticationFailed
from rest_framework.exceptions import status

from jsm_user_services import settings
from jsm_user_services.drf.exceptions import AuthenticationSkipped
from jsm_user_services.drf.exceptions import ExpiredToken
from jsm_user_services.drf.exceptions import InvalidToken
from jsm_user_services.drf.exceptions import NotAuthenticated
from jsm_user_services.drf.models import JsmUser
from jsm_user_services.drf.models import LoUser
from jsm_user_services.drf.models import LvSellerUser
from jsm_user_services.drf.models import LvUser
from jsm_user_services.services.user import current_jwt_token
from jsm_user_services.services.user import get_seller_user_data_from_server
from jsm_user_services.services.user import get_user_data_from_server
from jsm_user_services.support.auth_jwt import decode_jwt_token
from jsm_user_services.typings import Auth0TokenPayload
from jsm_user_services.typings import JsmRequest
from jsm_user_services.typings import LoTokenPayload
from jsm_user_services.typings import LoUserData
from jsm_user_services.typings import LvSellerData
from jsm_user_services.typings import LvTokenPayload
from jsm_user_services.typings import LvUserData

logger = logging.getLogger(__name__)

UserType = TypeVar("UserType", dict, AnonymousUser, JsmUser, LoUser, LvUser, LvSellerUser)
JwtPayloadType = TypeVar("JwtPayloadType", dict, Auth0TokenPayload, LoTokenPayload, LvTokenPayload)


class BaseJsmAuthentication(BaseAuthentication, Generic[JwtPayloadType, UserType]):
    """
    Base class for JWT authentication classes.

    Provides a method to retrieve user data from the request or from the user microservice.
    """

    user_claim: str = "sub"
    inject_user_data_to_request = settings.JSM_USER_SERVICES_DRF_APPEND_USER_DATA

    user_data: UserType
    user_id: str = "id"
    token: str = ""

    def __init__(self, *args, **kwargs) -> None:
        self.user_data = cast(UserType, {})

    def get_user_data_from_server(self) -> UserType:
        return cast(UserType, get_user_data_from_server())

    def get_user_data(self, request: JsmRequest) -> UserType:
        if self.user_data:
            return self.user_data

        try:
            user_data = self.get_user_data_from_server()
        except HTTPError as http_error:
            if http_error.response.status_code in (status.HTTP_401_UNAUTHORIZED, status.HTTP_403_FORBIDDEN):
                logger.debug("User not authorized in user service")
                raise AuthenticationFailed("User not authorized in user service") from http_error

            if http_error.response.status_code == status.HTTP_404_NOT_FOUND:
                logger.debug("User not found in user service")
                raise AuthenticationFailed("User not found") from http_error

            logger.exception("Failed to retrieve user data from server")
            raise AuthenticationFailed("Could not retrieve user data from server") from http_error

        if not isinstance(user_data, (dict, list)):
            logger.debug("User data is not a dictionary or list")
            raise AuthenticationFailed("User data is not a dictionary or list")

        if isinstance(user_data, dict):
            user_payload = user_data.get("data", user_data)
        else:
            user_payload = user_data[0]

        if self.inject_user_data_to_request:
            setattr(request, "jsm_user_data", user_payload)
            setattr(request, "jsm_user_id", user_payload[self.user_id])

        self.user_data = user_payload
        return self.user_data

    def custom_authenticate(self, request: JsmRequest) -> tuple[UserType, str] | None:
        raise NotImplementedError("Subclasses must implement the custom_authenticate method.")

    def _validate_token(self, token: str | None, user_claim: str | None = None) -> JwtPayloadType:
        if not token:
            raise NotAuthenticated()

        self.token = token

        try:
            payload = decode_jwt_token(token)
        except jwt.DecodeError as e:
            raise InvalidToken() from e
        except jwt.ExpiredSignatureError as e:
            raise ExpiredToken() from e

        user_claim = user_claim or self.user_claim
        if user_claim not in payload:
            logger.debug("User claim %s not present in jwt", self.user_claim)
            raise AuthenticationSkipped

        return cast(JwtPayloadType, payload)

    def validate_and_inject_token_payload(self, request: JsmRequest) -> JwtPayloadType:
        token = current_jwt_token()

        payload = self._validate_token(token)

        setattr(request, "jsm_token_payload", payload)

        return payload

    def authenticate(self, request: JsmRequest) -> tuple[UserType, str] | None:
        try:
            self.validate_and_inject_token_payload(request)
            self.get_user_data(request)
            return self.custom_authenticate(request)
        except AuthenticationSkipped:
            logger.debug("Authentication %s skipped", self.__class__.__name__)
            return None


class OauthJWTAuthentication(BaseJsmAuthentication[Auth0TokenPayload, LoUser]):
    """
    Authentication class for OAuth JWT tokens.
    Should be used in DRF views that need to use Auth0 tokens.

    user_claim is set just to be explicit.
    """

    user_id: str = "user_id_auth"

    def custom_authenticate(self, request: JsmRequest) -> tuple[LoUser, str] | None:
        return (LoUser(cast(LoUserData, self.user_data)), self.token)


class LoJWTAuthentication(BaseJsmAuthentication[LvTokenPayload, LoUser]):
    """
    Authentication class for LO JWT tokens.
    Should be used in DRF views that need to use LO tokens.

    user_claim is set just to be explicit.
    """

    user_claim: str = "jsm_identity"

    identity_user_claim: str = "uid"

    @override
    def validate_and_inject_token_payload(self, request: JsmRequest) -> LvTokenPayload:
        parent_payload = super().validate_and_inject_token_payload(request)

        # Workaround to allow legacy token to be used. Will be removed in the future.
        loyalty_token = self.token
        jsm_identity_token = cast(str | None, parent_payload.get("jsm_identity"))
        payload = self._validate_token(jsm_identity_token, self.identity_user_claim)
        self.token = loyalty_token

        setattr(request, "jsm_token_payload", payload)

        return payload

    def custom_authenticate(self, request: JsmRequest) -> tuple[LoUser, str] | None:
        return (LoUser(cast(LoUserData, self.user_data)), self.token)


class LvJWTAuthentication(BaseJsmAuthentication[LvTokenPayload, LvUser]):
    """
    Authentication class for LV JWT tokens.
    Should be used in DRF views that need to use LV tokens.
    """

    user_claim: str = "uid"

    def custom_authenticate(self, request: JsmRequest) -> tuple[LvUser, str] | None:
        return (LvUser(cast(LvUserData, self.user_data)), self.token)


class LVJWTSellerAuthentication(BaseJsmAuthentication[LvTokenPayload, LvSellerUser]):

    user_claim: str = "cid"

    def get_user_data_from_server(self):
        return get_seller_user_data_from_server()

    def custom_authenticate(self, request: JsmRequest) -> tuple[LvSellerUser, str] | None:
        return (LvSellerUser(cast(LvSellerData, self.user_data)), self.token)



"""
This file contains the permissions for the LV Django Apps.

Roles here are denormalized for easier use. Key being OUR name and values being the actual role name from identity.

Permissions ready to be used:

[JSM Admin]
- LVAdminJsmAdminPermission -> Admin app / Active User / Roles: JSMAdmin
- LVClientJsmAdminPermission -> Client app / Active User / Roles: JSMAdmin

[Buyer/Customer]
- LVClientBuyerPermission -> Client app / Active User / Roles: MasterClient

[Seller/Industry (Online and Offline) Admin]
- LVAdminSellerAdminPermission -> Admin app / Active User / Roles: MasterAdmin
- LVClientSellerAdminPermission -> Client app / Active User / Roles: MasterAdmin

[Seller/Industry (Online and Offline) API]
- LVAdminSellerApiPermission -> Admin app / Active User / Roles: Admin
- LVClientSellerApiPermission -> Client app / Active User / Roles: Admin

[Seller/Industry (Online and Offline) Coordinator]
- LVAdminSellerCoordinatorPermission -> Admin app / Active User / Roles: Coordinator
- LVClientSellerCoordinatorPermission -> Client app / Active User / Roles: Coordinator

[Seller/Industry (Online and Offline) Salesman]
- LVAdminSellerSalesmanPermission -> Admin app / Active User / Roles: Salesman
- LVClientSellerSalesmanPermission -> Client app / Active User / Roles: Salesman
"""

import logging

from typing import Any

from rest_framework.request import Request

from jsm_user_services.drf.permissions.base import JSMUserBasePermission
from jsm_user_services.enums import LVApplication
from jsm_user_services.enums import LVUserRole

logger = logging.LoggerAdapter(logging.getLogger(__name__), {"prefix": "JSM_USER_SERVICES"})


# Base permissions
class LVBasePermission(JSMUserBasePermission):
    """
    Base permission for LV Django Apps.
    """

    sources: set[LVApplication] = set()
    roles: set[LVUserRole] = set()

    _decoded_jwt_token: dict | None = None
    _decoded_jwt_token_keys: set[str] = set()
    _allowed_sources_set: set[str] = set()
    _allowed_roles_set: set[str] = set()

    _failed_validation: bool = False

    def __init__(self, *args: Any, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._allowed_sources_set = {source.value.lower() for source in self.sources}
        self._allowed_roles_set = {role.value.lower() for role in self.roles}

    def _is_valid(self, request: Request, key: str, value_to_check: set[str], **kwargs: Any) -> bool:
        if self._failed_validation:
            logger.debug("Validation already failed, skipping %s permission", key)
            return False

        logger.debug("Checking %s permission", key)
        decoded_token, _ = self.decode_jwt_token(request)

        if isinstance(decoded_token[key], list):
            token_set = {token_value.lower() for token_value in decoded_token[key]}
        else:
            token_set = {decoded_token[key].lower()}

        is_allowed_value = value_to_check.intersection(token_set)

        logger.debug("%s is allowed: %s", key, is_allowed_value)

        return bool(is_allowed_value)

    def has_permission(self, request: Request, view: Any) -> bool:
        logger.debug("Checking permissions")
        is_valid = super().has_permission(request, view)

        if is_valid and self.sources:
            is_valid = self._is_valid(request, "website", self._allowed_sources_set)
        if is_valid and self.roles:
            is_valid = self._is_valid(request, "role", self._allowed_roles_set)

        return is_valid


# Sources permissions
class LVAdminPermission(LVBasePermission):
    """
    Permission for LV Admin Apps for users with Source role.
    """

    sources: set[LVApplication] = {LVApplication.ADMIN_APPLICATION}


class LVClientPermission(LVBasePermission):
    """
    Permission for LV Client Apps for users with Source role.
    """

    sources: set[LVApplication] = {LVApplication.CLIENT_APPLICATION}


# JsmAdmin permissions
class LVAdminJsmAdminPermission(LVAdminPermission):
    """
    Permission for LV Admin Apps for users with JSM Admin role.
    """

    roles: set[LVUserRole] = {LVUserRole.JSM_ADMIN}


class LVClientJsmAdminPermission(LVClientPermission):
    """
    Permission for LV Client Apps for users with JSM Admin role.
    """

    roles: set[LVUserRole] = {LVUserRole.JSM_ADMIN}


# Buyer permissions
class LVClientBuyerPermission(LVClientPermission):
    """
    Permission for LV Admin Apps for users with Buyer role.
    """

    roles: set[LVUserRole] = {LVUserRole.BUYER}


# Seller (Industry) permissions
class LVAdminSellerAdminPermission(LVAdminPermission):
    """
    Permission for LV Admin Apps for users with Seller Admin role.
    """

    roles: set[LVUserRole] = {LVUserRole.SELLER_ADMIN}


class LVClientSellerAdminPermission(LVClientPermission):
    """
    Permission for LV Client Apps for users with Seller Admin role.
    """

    roles: set[LVUserRole] = {LVUserRole.SELLER_ADMIN}


# Seller API permissions
class LVAdminSellerApiPermission(LVAdminPermission):
    """
    Permission for LV Admin Apps for users with Seller API role.
    """

    roles: set[LVUserRole] = {LVUserRole.SELLER_API}


class LVClientSellerApiPermission(LVClientPermission):
    """
    Permission for LV Client Apps for users with Seller API role.
    """

    roles: set[LVUserRole] = {LVUserRole.SELLER_API}


# Seller Coordinator permissions
class LVAdminSellerCoordinatorPermission(LVAdminPermission):
    """
    Permission for LV Admin Apps for users with Seller Coordinator role.
    """

    roles: set[LVUserRole] = {LVUserRole.SELLER_COORDINATOR}


class LVClientSellerCoordinatorPermission(LVClientPermission):
    """
    Permission for LV Client Apps for users with Seller Coordinator role.
    """

    roles: set[LVUserRole] = {LVUserRole.SELLER_COORDINATOR}


# Seller Salesman permissions
class LVAdminSellerSalesmanPermission(LVAdminPermission):
    """
    Permission for LV Admin Apps for users with Seller Salesman role.
    """

    roles: set[LVUserRole] = {LVUserRole.SELLER_SALESMAN}


class LVClientSellerSalesmanPermission(LVClientPermission):
    """
    Permission for LV Client Apps for users with Seller Salesman role.
    """

    roles: set[LVUserRole] = {LVUserRole.SELLER_SALESMAN}



"""
Module for backwards compatibility with the old drf module.

Should be removed in the future.
"""

from jsm_user_services.drf import authentications
from jsm_user_services.drf import exceptions
from jsm_user_services.drf import helpers
from jsm_user_services.drf import permissions

__all__ = ["authentications", "exceptions", "helpers", "permissions"]



# Base helpers
from jsm_user_services.drf.permissions.base import protected
from jsm_user_services.drf.permissions.lo import ActiveUserPermission
from jsm_user_services.drf.permissions.lo import AnyLoggedUserPermission
from jsm_user_services.drf.permissions.lo import EmployeeOrManagerUserPermission
from jsm_user_services.drf.permissions.lo import EmployeeUserPermission
from jsm_user_services.drf.permissions.lo import GoogleRecaptchaPermission
from jsm_user_services.drf.permissions.lo import IndustrySellerUserPermission
from jsm_user_services.drf.permissions.lo import IsUserAllowedToPermission
from jsm_user_services.drf.permissions.lo import LoggedUserOauthAndLegacyPermission
from jsm_user_services.drf.permissions.lo import ManagerUserPermission
from jsm_user_services.drf.permissions.lo import OauthUserPermission
from jsm_user_services.drf.permissions.lo import OwnerOrManagerUserPermission
from jsm_user_services.drf.permissions.lo import OwnerUserPermission
from jsm_user_services.drf.permissions.lo import PendingValidationUserPermission
from jsm_user_services.drf.permissions.lo import RetailUserPermission
from jsm_user_services.drf.permissions.lo import RoleBasedPermission
from jsm_user_services.drf.permissions.lo import StatusBasedPermission
from jsm_user_services.drf.permissions.lo import UserBlockedByCpfAndBirthDateValidationFailedPermission
from jsm_user_services.drf.permissions.lv import LVAdminJsmAdminPermission
from jsm_user_services.drf.permissions.lv import LVAdminSellerAdminPermission
from jsm_user_services.drf.permissions.lv import LVAdminSellerApiPermission
from jsm_user_services.drf.permissions.lv import LVAdminSellerCoordinatorPermission
from jsm_user_services.drf.permissions.lv import LVAdminSellerSalesmanPermission
from jsm_user_services.drf.permissions.lv import LVBasePermission
from jsm_user_services.drf.permissions.lv import LVClientBuyerPermission
from jsm_user_services.drf.permissions.lv import LVClientJsmAdminPermission
from jsm_user_services.drf.permissions.lv import LVClientSellerAdminPermission
from jsm_user_services.drf.permissions.lv import LVClientSellerApiPermission
from jsm_user_services.drf.permissions.lv import LVClientSellerCoordinatorPermission
from jsm_user_services.drf.permissions.lv import LVClientSellerSalesmanPermission

__all__ = [
    # Base helpers
    "protected",
    # LO Permissions
    "AnyLoggedUserPermission",
    "EmployeeOrManagerUserPermission",
    "EmployeeUserPermission",
    "GoogleRecaptchaPermission",
    "IndustrySellerUserPermission",
    "IsUserAllowedToPermission",
    "LoggedUserOauthAndLegacyPermission",
    "ManagerUserPermission",
    "OauthUserPermission",
    "OwnerOrManagerUserPermission",
    "OwnerUserPermission",
    "RetailUserPermission",
    "ActiveUserPermission",
    "PendingValidationUserPermission",
    "RoleBasedPermission",
    "StatusBasedPermission",
    "UserBlockedByCpfAndBirthDateValidationFailedPermission",
    # LV Permissions
    "protected",
    "LVBasePermission",
    "LVAdminJsmAdminPermission",
    "LVClientJsmAdminPermission",
    "LVClientBuyerPermission",
    "LVAdminSellerAdminPermission",
    "LVClientSellerAdminPermission",
    "LVAdminSellerApiPermission",
    "LVClientSellerApiPermission",
    "LVAdminSellerCoordinatorPermission",
    "LVClientSellerCoordinatorPermission",
    "LVAdminSellerSalesmanPermission",
    "LVClientSellerSalesmanPermission",
]



from enum import Enum


# LV Enums
class LVApplication(Enum):
    CLIENT_APPLICATION = "ClientApplication"  # Public application, used by customers.
    ADMIN_APPLICATION = "AdminApplication"  # Admin application, used by admins.


class LVUserRole(Enum):
    # Seller and industry terms are interchangeable in this context.
    JSM_ADMIN = "JSMAdmin"  # JSM Full admin role, used by JSM users. Used to manage everything in the system.
    BUYER = "MasterClient"  # Customer role, used by customers. This is the default role for customers.
    SELLER_ADMIN = "MasterAdmin"  # Seller Admin role, used by sellers. Used to manage seller data in admin.
    SELLER_API = "Admin"  # API role. Used by sellers to access the API.
    SELLER_COORDINATOR = (
        "Coordinator"  # Coordinator role, used by sellers to manage some things in the seller in admin.
    )
    CUSTOMER = "Customer"  # Customer role, used mainly by loyalty jwt.
    SELLER_SALESMAN = "Salesman"  # Salesman role, used by sellers to sell products.



import logging

from requests.exceptions import HTTPError
from rest_framework import permissions
from rest_framework.request import Request
from rest_framework.views import APIView

from jsm_user_services import settings
from jsm_user_services.services.user import get_user_data_from_server

logger = logging.getLogger(__name__)


class JSMUserBasePermission(permissions.BasePermission):
    """
    Base class for JSM user permissions. Implements methods for validating requests against
    the user micro service so that any type of permission can be carried out (role, status, etc).
    """

    APPEND_USER_DATA = settings.JSM_USER_SERVICES_DRF_APPEND_USER_DATA
    USER_DATA_ATTR_NAME = settings.JSM_USER_SERVICES_DRF_REQUEST_USER_DATA_ATTR_NAME
    USER_SERVICE_NOT_AUTHORIZED_CODES = (401, 403, 404)

    _decoded_jwt_token: dict | None = None
    _decoded_jwt_token_keys: set[str] | None = None

    _protected: bool = False
    _forbidden: bool = False
    _protected_keys: set[str] | None = None

    @classmethod
    def _retrieve_user_data(cls, request: Request) -> dict:
        try:
            user_data = getattr(request, cls.USER_DATA_ATTR_NAME)
        except AttributeError:
            user_data = get_user_data_from_server()

        return user_data

    @classmethod
    def _validate_request_against_user_service(cls, request: Request, append_user_data_to_request: bool = True) -> dict:
        """
        Gets valid user_data from the User micro service.
        """
        if not settings.JSM_USER_SERVICE_REQUEST_USER_DATA:
            return {}

        try:
            user_data: dict = cls._retrieve_user_data(request)

            if append_user_data_to_request:
                setattr(request, cls.USER_DATA_ATTR_NAME, user_data)

        except HTTPError as e:
            if e.response.status_code in cls.USER_SERVICE_NOT_AUTHORIZED_CODES:
                return {}

            raise

        return user_data

    def decode_jwt_token(self, request: Request) -> tuple[dict, set[str]]:
        if self._decoded_jwt_token and self._decoded_jwt_token_keys:
            return self._decoded_jwt_token, self._decoded_jwt_token_keys

        decoded_token = getattr(request, "jsm_token_payload", None)
        logger.debug("Decoded token: %s", decoded_token)

        if not decoded_token:
            raise ValueError("Decoded token is not present in the request")

        self._decoded_jwt_token = decoded_token
        self._decoded_jwt_token_keys = {key.lower() for key in decoded_token.keys()}
        return decoded_token, self._decoded_jwt_token_keys

    def _protect(self, request: Request) -> bool:
        setattr(request, "is_protected", self._protected)
        setattr(request, "protected_keys", self._protected_keys or set())
        setattr(request, "should_return_forbidden", self._forbidden)

        _, decoded_token_keys = self.decode_jwt_token(request)
        if self._forbidden and decoded_token_keys and "fid" in decoded_token_keys:
            return False

        return True

    def has_permission(self, request: Request, view: APIView) -> bool:
        """
        Abstract method that must be implemented by children classes.
        """
        is_valid = True

        if self._protected:
            is_valid = self._protect(request)

        return is_valid


def protected(
    cls: type["JSMUserBasePermission"], keys: set[str] | None = None, should_return_forbidden: bool = False
) -> type["JSMUserBasePermission"]:
    """
    Decorator to mark a permission as protected.
    """
    cls._protected = True
    cls._forbidden = should_return_forbidden
    cls._protected_keys = keys or set()
    return cls



from typing import List
from typing import Literal
from typing import NotRequired
from typing import Optional
from typing import Protocol
from typing import TypedDict
from uuid import UUID

from jsm_user_services.enums import LVApplication
from jsm_user_services.enums import LVUserRole


# Jsm Request Protocol
class JsmRequest(Protocol):
    jsm_token_payload: dict
    jsm_user_id: str
    jsm_user_data: dict


class Claim(TypedDict):
    key: str
    value: str


class BaseUserData(TypedDict):
    id: str
    name: str


# Auth0 Token Payload
class Auth0TokenPayload(TypedDict):
    client_id: str  # Client ID
    session_id: str  # Session ID
    user_id_auth: str  # User ID Auth, usually means that this token is being impersonated.
    iss: str  # Issuer
    sub: str  # Subject
    aud: list[str]  # Audience
    iat: int  # Issued at
    exp: int  # Expiration time
    scope: str  # Scope, separated by spaces
    azp: str  # Authorized party


# Lv Token Payload
class LvTokenPayload(TypedDict):
    email: str
    unique_name: str
    website: LVApplication
    uid: str | UUID
    role: LVUserRole
    nbf: int  # Token not valid before time
    exp: int  # Expiration time
    iat: int  # Issued at
    iss: str  # Issuer
    aud: str  # Audience
    fid: NotRequired[str | UUID | None]  # Federated ID
    sellers: NotRequired[str | UUID | list[str | UUID] | None]  # Sellers ID
    cnpjs: NotRequired[str | list[str] | None]  # CNPJs


# Lo Token Payload
class LoTokenPayload(TypedDict):
    """
    Legacy token payload for LO. This token is a wrapper for Lv Token Payload.
    If needed, use the info inside jsm_identity.
    """

    token_type: Literal["access"]
    exp: int  # Expiration time
    iat: int  # Issued at
    jti: str  # JWT ID
    jsm_identity: str  # Lv Token Payload


# Lo User Data
class Email(TypedDict):
    type: str
    email: str


class Phone(TypedDict):
    type: str
    number: str


class BlockedReason(TypedDict):
    reason: str
    blocked_date: str


class LoUserData(BaseUserData):
    cpf: str
    roles: List[str]
    emails: List[Email]
    gender: str
    phones: List[Phone]
    status: str
    mediums: NotRequired[List[str]]
    birthday: str
    username: str
    addresses: NotRequired[List[dict]]
    occupation: str
    blocked_reason: NotRequired[BlockedReason | None]
    not_allowed_to: NotRequired[List[str]]
    registered_date: str
    industry_register_helper: NotRequired[str]
    segments: List[str] | None
    user_id_auth: str | None


# Lv User Data
class LvUserData(BaseUserData):
    email: str
    cnpj: str
    cpf: str
    gender: int
    cep: NotRequired[Optional[str]]
    street: NotRequired[Optional[str]]
    complement: NotRequired[Optional[str]]
    phoneNumber: NotRequired[Optional[str]]
    neighbourhood: NotRequired[Optional[str]]
    state: int
    city: NotRequired[Optional[str]]
    mobile: NotRequired[Optional[str]]
    number: NotRequired[Optional[str]]
    twoFactorEnabled: bool
    uf: str
    metadata: NotRequired[Optional[str]]
    claims: List[Claim]
    roles: List[str]
    sellers: List[str]
    permissions: List[str]
    cnpjs: List[str]
    cpfs: List[str]
    fidRoles: NotRequired[Optional[List[str]]]
    isJSMAdmin: bool


class LvUserDataResponse(TypedDict):
    data: LvUserData
    success: bool


class LvSellerData(BaseUserData):
    """Seller user data for LV JWT authentication."""

    seller_id: str



from django.conf import settings

from jsm_user_services.exception import MissingRequiredConfiguration


def get_settings_or_raise_missing_config(
    settings_key: str, default: str | None = None, cast: type = str, required: bool = True
):
    value = getattr(settings, settings_key, default)

    if value is None and required:
        raise MissingRequiredConfiguration(f"The variable {settings_key} is missing")
    elif value is None:
        return default

    return cast(value)


# User Service related settings
USER_API_HOST = get_settings_or_raise_missing_config("USER_API_HOST")
USER_API_TOKEN = get_settings_or_raise_missing_config("USER_API_TOKEN")
USER_API_PROFILE_ENDPOINT = get_settings_or_raise_missing_config("USER_API_PROFILE_ENDPOINT", "/v1/users/me/")
USER_API_SELLER_PROFILE_ENDPOINT = get_settings_or_raise_missing_config(
    "USER_API_SELLER_PROFILE_ENDPOINT", "/v1/me/sellers/"
)
JSM_USER_SERVICE_REQUEST_USER_DATA = get_settings_or_raise_missing_config(
    "JSM_USER_SERVICE_REQUEST_USER_DATA", "True", cast=bool
)
JSM_USER_SERVICE_HTTP_TIMEOUT = get_settings_or_raise_missing_config("JSM_USER_SERVICE_HTTP_TIMEOUT", "30", cast=float)
JSM_USER_SERVICES_DRF_APPEND_USER_DATA = get_settings_or_raise_missing_config(
    "JSM_USER_SERVICES_DRF_APPEND_USER_DATA", "True", cast=bool
)
JSM_USER_SERVICES_DRF_REQUEST_USER_DATA_ATTR_NAME = get_settings_or_raise_missing_config(
    "JSM_USER_SERVICES_DRF_REQUEST_USER_DATA_ATTR_NAME", "jsm_user_data"
)
REQUEST_ID_CONFIG = get_settings_or_raise_missing_config("REQUEST_ID_CONFIG", required=False, cast=dict)

# JWT related settings
JSM_JWT_ALGORITHM = get_settings_or_raise_missing_config("JSM_JWT_ALGORITHM", "HS256")
JSM_JWT_DECODE = get_settings_or_raise_missing_config("JSM_JWT_DECODE", "True", cast=bool)
JSM_JWT_SHOULD_VERIFY_SIGNATURE = get_settings_or_raise_missing_config(
    "JSM_JWT_SHOULD_VERIFY_SIGNATURE", "False", cast=bool
)
JSM_JWT_SECRET_KEY = get_settings_or_raise_missing_config("JSM_JWT_SECRET_KEY", required=False)
JSM_JWT_AUDIENCE = get_settings_or_raise_missing_config("JSM_JWT_AUDIENCE", default="audience")

# Google Recaptcha related settings
GOOGLE_RECAPTCHA_URL = get_settings_or_raise_missing_config("GOOGLE_RECAPTCHA_URL", required=False)
GOOGLE_RECAPTCHA_SECRET_KEY = get_settings_or_raise_missing_config("GOOGLE_RECAPTCHA_SECRET_KEY", required=False)
GOOGLE_RECAPTCHA_SCORE_THRESHOLD = get_settings_or_raise_missing_config(
    "GOOGLE_RECAPTCHA_SCORE_THRESHOLD", required=False
)
GOOGLE_RECAPTCHA_BYPASS_IP_HEADER_NAME = get_settings_or_raise_missing_config(
    "GOOGLE_RECAPTCHA_BYPASS_IP_HEADER_NAME", required=False
)

# Everest related settings
EVEREST_API_KEY = get_settings_or_raise_missing_config("EVEREST_API_KEY", required=False)
EVEREST_API_HOST = get_settings_or_raise_missing_config("EVEREST_API_HOST", required=False)
EVEREST_LIST_INVALIDS_STATUS_RESPONSES = get_settings_or_raise_missing_config(
    "EVEREST_LIST_INVALIDS_STATUS_RESPONSES", required=False
)



"""
This module contains the ephemeral models for user types.

It is named this way to follow Django conventions set by AnonymousUser.
"""

from django.contrib.auth.models import AnonymousUser

from jsm_user_services.typings import BaseUserData
from jsm_user_services.typings import LoUserData
from jsm_user_services.typings import LvSellerData
from jsm_user_services.typings import LvUserData


class JsmUser(AnonymousUser):
    id = None
    pk = None
    username = ""
    is_staff = False  # It will always be False
    is_active = False  # It will always be False
    is_superuser = False  # It will always be False

    user_data: BaseUserData

    def __init__(self, *args, **kwargs) -> None:
        self.set_fields()

    def set_fields(self) -> None:
        if not self.user_data:
            return

        self.id = self.user_data["id"]
        self.pk = self.user_data["id"]
        self.username = self.user_data["name"]

    @property
    def is_authenticated(self) -> bool:
        """
        Always returns True as the Jsm User is only set if the authentication succeeds.
        """
        return True


class LoUser(JsmUser):
    user_data: LoUserData

    def __init__(self, user_data: LoUserData, *args, **kwargs) -> None:
        self.user_data = user_data
        super().__init__(*args, **kwargs)


class LvUser(JsmUser):
    user_data: LvUserData

    def __init__(self, user_data: LvUserData, *args, **kwargs) -> None:
        self.user_data = user_data
        super().__init__(*args, **kwargs)


class LvSellerUser(JsmUser):
    user_data: LvSellerData

    def __init__(self, user_data: LvSellerData, *args, **kwargs) -> None:
        self.user_data = user_data
        super().__init__(*args, **kwargs)



import re

from typing import Optional

import jwt

from jsm_user_services import settings


def decode_jwt_token(jwt_token: str) -> dict:
    """
    Gets a decoded JWT token with the HS256 algorithm.
    """
    if not settings.JSM_JWT_DECODE:
        return {}

    options = {"verify_exp": True}
    kwargs = {}

    if settings.JSM_JWT_ALGORITHM:
        kwargs["algorithms"] = [settings.JSM_JWT_ALGORITHM]

    if settings.JSM_JWT_AUDIENCE:
        kwargs["audience"] = settings.JSM_JWT_AUDIENCE

    if not settings.JSM_JWT_SHOULD_VERIFY_SIGNATURE:
        options["verify_signature"] = False
    else:
        options["verify_signature"] = True
        kwargs["key"] = settings.JSM_JWT_SECRET_KEY

    return jwt.decode(jwt_token, options=options, **kwargs)


def get_bearer_authorization_token(authorization_value: str) -> Optional[str]:
    """
    Retrieve a bearer authorization token from an Authorization header value.

    It expects the header value to be something on the lines of: "Bearer token".

    Examples:
    - get_bearer_authorization_token("Bearer token") # returns "token"
    - get_bearer_authorization_token("bearer token") # returns None
    - get_bearer_authorization_token("Token token") # returns None
    - get_bearer_authorization_token("whatever") # returns None
    """
    match = re.match("Bearer", authorization_value)

    if not match:
        return None

    auth_type_beginning = match.span()[1]
    jwt_token = authorization_value[auth_type_beginning:].strip()

    return jwt_token



import logging

from jsm_user_services.support.http_utils import request
from jsm_user_services.support.local_threading_utils import get_from_local_threading
from jsm_user_services.support.settings_utils import get_from_settings_or_raise_missing_config

logger = logging.getLogger(__name__)


def perform_recaptcha_validation(g_recaptcha_response: str) -> bool:
    """
    Performs a request to Google in order to validate the reCAPTCHA.
    For more details, check: https://developers.google.com/recaptcha/docs/verify
    """

    google_recaptcha_url = get_from_settings_or_raise_missing_config(
        "GOOGLE_RECAPTCHA_URL", "https://www.google.com/recaptcha/api/siteverify"
    )
    google_recaptcha_secret_key = get_from_settings_or_raise_missing_config("GOOGLE_RECAPTCHA_SECRET_KEY")

    min_score_threshold = float(get_from_settings_or_raise_missing_config("GOOGLE_RECAPTCHA_SCORE_THRESHOLD", "0.8"))

    google_recaptcha_bypass_ip_header_name = get_from_settings_or_raise_missing_config(
        "GOOGLE_RECAPTCHA_BYPASS_IP_HEADER_NAME", "true-client-ip"
    )

    user_ip = get_from_local_threading("user_ip")
    headers = {google_recaptcha_bypass_ip_header_name: user_ip}
    data = {"response": g_recaptcha_response, "secret": google_recaptcha_secret_key}

    if user_ip:
        data["remoteip"] = user_ip

    logger.debug("Performing request to Google to check if recaptcha is valid")

    with request() as r:
        resp = r.post(google_recaptcha_url, data=data, headers=headers)

    result_json = resp.json()

    if resp.status_code != 200:
        logger.warning(
            "[GoogleRecaptcha] Validation failed for response: %s | Status code: %s | Response body: %s",
            g_recaptcha_response,
            resp.status_code,
            result_json,
        )
        return False

    if result_json.get("success") is not True:
        logger.info("[GoogleRecaptcha] Recaptcha is not valid: %s", result_json)
        return False

    if result_json.get("score", 0) < min_score_threshold:
        logger.info("[GoogleRecaptcha] Score %s below threshold %s", result_json.get("score", 0), min_score_threshold)
        return False

    logger.info(
        "[GoogleRecaptcha] Validation succeeded. min_score_threshold: %s, response body: %s",
        min_score_threshold,
        result_json,
    )
    return True
